/*
 * $Id: mktime.c,v 2.3 1999/10/21 02:49:26 john Exp $
 *
 * Command line utility with the functionality of mktime() (plus some).
 */

#include "config.h"
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <errno.h>
#include <assert.h>
#if HAVE_UNISTD_H
# include <unistd.h>
#endif
#if HAVE_STDLIB_H
# include <stdlib.h>
#endif
#ifdef HAVE_LOCALE_H
# include <locale.h>
#endif
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#if HAVE_GETOPT_DECLS == 0
extern char	*optarg;
extern int	optind, opterr;
#endif

#define	OPTIONS	"H:M:S:y:m:d:t:F:fT:D:z"
#define USAGE	"Usage: %s [-f] [-F fmt] [-T hh:mm[:ss]] [-H hour] [-M min] [-S sec]\n" \
		"       [-D [yyyy-]mm-dd] [-y year] [-m mon] [-d day] [-t time_t] [-z]\n"

#if HAVE_EXIT_STATUSES
# define EX_FAIL_TIME	1	/* Bad time */
# define EX_FAIL_USAGE	2	/* Usage error */
# define EX_FAIL_RUN	3	/* Some sort of runtime failure */
#else
# ifndef EXIT_FAILURE
#  define EXIT_FAILURE	1
# endif
# define EX_FAIL_TIME	EXIT_FAILURE
# define EX_FAIL_USAGE	EXIT_FAILURE
# define EX_FAIL_RUN	EXIT_FAILURE
#endif

#define DEF_BUFLEN	1024	/* Big enough for both a printed time_t, and
				 * typical formatted output */

#define TIME_T_FMT_CHAR	't'	/* Format character to be replaced by time_t
				 * value */

#define DEF_STRFTIME_FMT	"%Y-%m-%d %H:%M:%S"	/* ISO 8601:1988 */

/*
 * Ugly stuff for dealing with time_t values.
 */

#if HAVE_LONG_DOUBLE_TIME_T
# ifndef HAVE_FP_TIME_T
#  define HAVE_FP_TIME_T	1
# endif
#else
# if HAVE_LONG_LONG_TIME_T
#  undef HAVE_FP_TIME_T
# endif
#endif

#if HAVE_FP_TIME_T
# if HAVE_LONG_DOUBLE_TIME_T
#  define TIME_T_TYPE	long double	/* Type to use */
#  ifndef TIME_T_FMT
#   define TIME_T_FMT	"%.128Lg"	/* Format for printing */
#  endif
#  define TIME_T_SCAN	"%Lf"		/* Format for scanning */
# else
#  define TIME_T_TYPE	double		/* Type to use */
#  ifndef TIME_T_FMT
#   define TIME_T_FMT	"%.128g"	/* Format for printing */
#  endif
# endif
#else
# if HAVE_LONG_LONG_TIME_T
#  define TIME_T_TYPE	long long	/* Type to use */
#  ifndef TIME_T_FMT
#   define TIME_T_FMT	"%Ld"		/* Format for printing */
#  endif
#  ifndef TIME_T_SCAN
#   define TIME_T_SCAN	"%Ld"		/* Format for scanning */
#  endif
# else
#  define TIME_T_TYPE	long		/* Type to use */
#  define TIME_T_FMT	"%ld"		/* Format for printing */
# endif
#endif

typedef enum {
	VALUE_ADJUST,
	VALUE_SET
} value_type;

struct value {
	value_type	type;
	int		value;
};

struct date_time {
	struct value	sec;
	struct value	min;
	struct value	hour;
	struct value	mday;
	struct value	mon;
	struct value	year;
} arg = {
	{ VALUE_ADJUST, 0 },
	{ VALUE_ADJUST, 0 },
	{ VALUE_ADJUST, 0 },
	{ VALUE_ADJUST, 0 },
	{ VALUE_ADJUST, 0 },
	{ VALUE_ADJUST, 0 }
};

char	*progname;

static char	RCSid[] = "$Id: mktime.c,v 2.3 1999/10/21 02:49:26 john Exp $";

static time_t	get_time_t(const char *);
static void	get_value(const char *, struct value *);
static int	adjust(int, struct value *);
static size_t	xstrftime(char *, size_t, const char *, const struct tm *,
			time_t);

/*
 * Get a time_t value.  This function tries to use an appropriate
 * mechanism to read a time_t, but if they are `long long' it uses a
 * non-standard sscanf() format, and for long double it uses sscanf() so
 * the error checking on these two is not as good.  It could also
 * conceivably overflow a time_t, but there's not good way to check for
 * this.
 */

static time_t
get_time_t(const char *str)
{
	TIME_T_TYPE	result;
	char		*end;		/* end pointer for strtoX() */

#if HAVE_LONG_LONG_TIME_T || HAVE_LONG_DOUBLE_TIME_T
	if (sscanf(str, TIME_T_SCAN, &result) != 1) {
		fprintf(stderr, "%s: invalid time_t: %s\n", progname, str);
		exit(EX_FAIL_USAGE);
	}
#else
	errno = 0;
# if HAVE_FP_TIME_T
	result = strtod(str, &end);
# else
	result = strtol(str, &end, 10);
# endif
	if (errno == ERANGE || *end != '\0') {
		fprintf(stderr, "%s: invalid time_t: %s\n", progname, str);
		exit(EX_FAIL_USAGE);
	}
#endif
	return (time_t)result;
}

/*
 * Get the value associated with an option.  If the value has an
 * explicit sign, we adjust the current value, otherwise, it is of type
 * VALUE_SET and we set the value.
 */
static void
get_value(const char *str, struct value *vp)
{
	long			result;
	char			*end;		/* end pointer for strtol() */
	const unsigned char	*s;		/* used to look for explicit
						 * sign on option value */

	errno = 0;
	result = strtol(str, &end, 10);
	if ((errno == ERANGE && (result == LONG_MAX || result == LONG_MIN)) ||
	    *end != '\0') {
		fprintf(stderr, "%s: invalid value: %s\n", progname, str);
		exit(EX_FAIL_USAGE);
	} else {
		s = (const unsigned char *)str;
		while (isspace(*s)) {
			++s;
		}
		if (*s == '+' || *s == '-') {
			vp->value += result;
		} else {
			vp->type = VALUE_SET;
			vp->value = result;
		}
	}
}

/*
 * Get a time in the format hh:mm[:ss].  Checking is not particularly
 * extensive, particularly for trailing garbage.
 */

static void
get_time(const char *str, struct date_time *argp)
{
	unsigned int	hh, mm, ss;
	int		n;

	n = sscanf(str, "%u:%u:%u", &hh, &mm, &ss);
	if (n == 2) {
		ss = 0;
	}
	if (n < 2 || n > 3 || hh > INT_MAX || mm > INT_MAX || ss > INT_MAX) {
		fprintf(stderr, "%s: invalid time: %s\n", progname, str);
		exit(EX_FAIL_USAGE);
	}
	argp->hour.type = VALUE_SET;
	argp->hour.value = (int)hh;
	argp->min.type = VALUE_SET;
	argp->min.value = (int)mm;
	argp->sec.type = VALUE_SET;
	argp->sec.value = (int)ss;
}

/*
 * Get a date in the format [yyyy-]mm-dd.  Checking is not particularly
 * extensive, particularly for trailing garbage.
 */

static void
get_date(const char *str, struct date_time *argp)
{
	unsigned int	yy, mm, dd;
	int		n;

	n = sscanf(str, "%u-%u-%u", &yy, &mm, &dd);
	if (n == 2) {
		dd = mm;
		mm = yy;
		yy = 0;
	}
	if (n < 2 || n > 3 || yy > INT_MAX || mm > INT_MAX || dd > INT_MAX) {
		fprintf(stderr, "%s: invalid time: %s\n", progname, str);
		exit(EX_FAIL_USAGE);
	}
	if (n == 3) {
		argp->year.type = VALUE_SET;
		argp->year.value = (int)yy;
	}
	argp->mon.type = VALUE_SET;
	argp->mon.value = (int)mm;
	argp->mday.type = VALUE_SET;
	argp->mday.value = (int)dd;
}

/*
 * Either adjust or set a value
 */

static int
adjust(int num, struct value *vp)
{
	if (vp->type == VALUE_ADJUST) {
		/* Could overflow */
		num += vp->value;
	} else {
		num = vp->value;
	}

	return num;
}

/*
 * This is essentially a strftime() ``pre-processor'' that looks for the
 * time_t format char, and replaces it with the appropriate time_t value
 */

static size_t
xstrftime(char *outbuf, size_t maxlen, const char *fmt, const struct tm *tp,
	  time_t t)
{
	char		timestr[DEF_BUFLEN];	/* time_t value as a string */
	int		timelen;		/* length of time_t value */
	char		*newfmt;		/* expanded format string */
	const char	*src;
	char		*dest;
	size_t		result;

	/*
	 * Make a guess as to how long the new format string will be.  At
	 * worst case, it will be full of just time formats, so it will
	 * require (len + 1) / 2 times the length of the printed time_t
	 * value, plus one for the trailing \0.
	 *
	 * sprintf() should return the length, but just in case this is
	 * some ancient one, we'll do it seperately.
	 */
	(void) sprintf(timestr, TIME_T_FMT, (TIME_T_TYPE)t);
	timelen = strlen(timestr);
	assert(timelen < sizeof(timestr));
	newfmt = (char *)malloc((strlen(fmt) + 1) / 2 * timelen + 1);
	if (newfmt == NULL) {
		fprintf(stderr, "%s: out of memory\n", progname);
		exit(EX_FAIL_RUN);
	}

	/*
	 * Craft a new format string based on the old one, with the time
	 * value substituted in.
	 */
	src = fmt;
	dest = newfmt;

	while (*src != '\0') {
		if (src[0] == '%' && src[1] == TIME_T_FMT_CHAR) {
			strcpy(dest, timestr);
			dest += timelen;
			src += 2;
		} else {
			*dest++ = *src++;
		}
	}
	*dest = '\0';

	/*
	 * Now call strftime()
	 */

	result = strftime(outbuf, maxlen, newfmt, tp);

	free(newfmt);

	return result;
}

int
main(int argc, char *argv[])
{
	time_t		basetime = -1;		/* base time for adjustments */
	time_t		answer;			/* result from mktime() */
	struct tm	newtm;			/* arg for mktime() */
	struct tm	*pt;			/* result to be printed */
	int		c;			/* option character */
	const char	*format = DEF_STRFTIME_FMT; /* output time format */
	int		fix = 1;		/* try to fix bad times */
	char		*buffer = NULL;		/* buffer for output str */
	size_t		buflen = DEF_BUFLEN;	/* length of buffer */
	size_t		len;			/* result from strftime */

	/*
	 * Set locale for appropriate printing of the output time
	 */
#if HAVE_SETLOCALE
	(void) setlocale(LC_TIME, "");
#endif

	/*
	 * Save program name for failure messages
	 */
	progname = strrchr(argv[0], '/');
	if (progname != NULL) {
		++progname;
	} else {
		progname = argv[0];
	}

	/*
	 * Process arguments
	 */
	while ((c = getopt(argc, argv, OPTIONS)) != EOF) {
		switch (c) {
		case 'f':
			fix = 0;
			break;
		case 'F':
			format = optarg;
			break;
		case 't':
			basetime = get_time_t(optarg);
			break;
		case 'T':
			get_time(optarg, &arg);
			break;
		case 'H':
			get_value(optarg, &arg.hour);
			break;
		case 'M':
			get_value(optarg, &arg.min);
			break;
		case 'S':
			get_value(optarg, &arg.sec);
			break;
		case 'D':
			get_date(optarg, &arg);
			break;
		case 'y':
			get_value(optarg, &arg.year);
			break;
		case 'm':
			get_value(optarg, &arg.mon);
			break;
		case 'd':
			get_value(optarg, &arg.mday);
			break;
		case 'z':
			arg.hour.type = VALUE_SET;
			arg.hour.value = 0;
			arg.min.type = VALUE_SET;
			arg.min.value = 0;
			arg.sec.type = VALUE_SET;
			arg.sec.value = 0;
			break;
		default:
			fprintf(stderr, USAGE, progname);
			exit(EX_FAIL_USAGE);
			break;
		}
	}

	/*
	 * Get current time if no base time was specified.
	 */
	if (basetime == (time_t)-1) {
		basetime = time(NULL);
		if (basetime == (time_t)-1) {
			fprintf(stderr, "%s: current time unavailable\n", progname);
			exit(EX_FAIL_RUN);
		}
	}

	/*
	 * Fill in the struct tm with the default time, and adjust it
	 * according to the specified options.
	 */

	newtm = *localtime(&basetime);
	newtm.tm_isdst = -1;
	newtm.tm_sec = adjust(newtm.tm_sec, &arg.sec);
	newtm.tm_min = adjust(newtm.tm_min, &arg.min);
	newtm.tm_hour = adjust(newtm.tm_hour, &arg.hour);
	newtm.tm_mday = adjust(newtm.tm_mday, &arg.mday);
	if (arg.mon.type == VALUE_SET) {
		/* Months need to be zero-based */
		--arg.mon.value;
	}
	newtm.tm_mon = adjust(newtm.tm_mon, &arg.mon);
	if (arg.year.type == VALUE_SET && arg.year.value >= 1900) {
		/* Year needs to be years from 1900 */
		arg.year.value -= 1900;
	}
	newtm.tm_year = adjust(newtm.tm_year, &arg.year);

	/*
	 * Call mktime() to get the answer time
	 */
	answer = mktime(&newtm);
	if (answer == (time_t)-1 && fix) {
		newtm.tm_isdst = 0;
		answer = mktime(&newtm);
		if (answer == (time_t)-1) {
			/* One last try... */
			newtm.tm_isdst = 1;
			answer = mktime(&newtm);
		}
	}
	if (answer == (time_t)-1) {
		fprintf(stderr, "%s: invalid time\n", progname);
		exit(EX_FAIL_TIME);
	}

	/*
	 * Print the result, successively trying larger buffers
	 * if the result won't fit.
	 */
	pt = localtime(&answer);
	buffer = (char *)malloc(buflen);
	do {
		if (buffer == NULL) {
			fprintf(stderr, "%s: out of memory for result (possible format string error)\n", progname);
			exit(EX_FAIL_RUN);
		}
		len = xstrftime(buffer, buflen, format, pt, answer);
		if (len == 0) {
			free(buffer);
			buflen *= 2;
			buffer = (char *)malloc(buflen);
		}
	} while (len == 0);

	puts(buffer);
	free(buffer);

	exit(EXIT_SUCCESS);
}
